classdef PolyRef < matlab.mixin.Copyable
    %POLYREF uses Matlab B-splines (piecewise polynomial function) to
    %represent transitions in a variable between t=[t0,t1].
    %This class outputs a trajectory in time and its first two time
    %derivatives
    
    properties  (SetAccess=private, GetAccess=private)
        t0; % transition start time
        t1; % transition end time
        
        % polynomial trajectory calculated by numerically inverting system of equations
        p_y; % reference to position polynomial trajectory 
        p_dy; % reference to velocity polynomial trajectory 
        p_ddy; % reference to acceleration polynomial trajectory 
        
        % spline trajectory calculated using spapi tool in curve fitting toolbox
        n=6; % spline order (a spline is a piecewise polynomial function of degree n-1)
        ps_y; % reference to position spline trajectory
        ps_dy; % reference to velocity spline trajectory
        ps_ddy; % reference to acceleration spline trajectory        
    end
    
    properties  (SetAccess=private, GetAccess=public)
        y;  % position
        dy; % velocity
        ddy;% acceleration
    end
    
    methods
        function obj = PolyRef(t0,t1,y0,y1,dy0,dy1,ddy0,ddy1)
            %POLYREF Construct an instance of this class to evaluate the polynomial trajectory parameters arguments:
            %   t0: transition start time
            %   t1: transition end time
            %   y0: position at t0
            %   y1: position at t1
            %   dy0: velocity at t0
            %   dy1: velocity at t1
            %   ddy0:   acceleration at t0
            %   ddy1:   acceleration at t1
            if  nargin > 0
                obj.t0=t0;
                obj.t1=t1;

                % TASK 2: calculate polynomial coefficients by numerically
                % inverting system of equations
                T=t1-t0;
                S=[1 1 1;5 4 3;20 12 6];
                obj.p_y=[diag([1/T^5,1/T^4,1/T^3])*(S\[y1-y0-T*dy0-1/2*ddy0*T^2;T*(dy1-dy0-T*ddy0);T^2*(ddy1-ddy0)]);1/2*ddy0;dy0;y0];

                obj.p_dy=polyder(obj.p_y);
                obj.p_ddy=polyder(obj.p_dy);

                % TASK 3: calculate polynomial coefficents using spapi tool from curve fitting toolbox
                obj.ps_y = spapi(obj.n,[t0 t0 t0 t1 t1 t1],[y0 dy0 ddy0 y1 dy1 ddy1]);
                obj.ps_dy = fnder(obj.ps_y);
                obj.ps_ddy = fnder(obj.ps_y,2);
            end        
        end
        
        function obj = eval(obj,tt)
            %EVAL evaluates the trajectory defined in for the times in vector tt and saves the output in public properties y,dy,ddy. 
            
            % TASK 2: evaluate the polynomial and its first 2 time derivatives
            
            t=tt;
          
            obj.y = polyval(obj.p_y,t);
            obj.dy = polyval(obj.p_dy,t);
            obj.ddy = polyval(obj.p_ddy,t);

        end
        
        function obj = eval_spline(obj,tt)
            %EVAL_SPLINE evaluates the trajectory defined by for the times in vector tt and saves the output in public properties y,dy,ddy. 
            t=tt;

            % TASK 3: evaluate the spline-based polynomial and its first 2 time derivatives
            obj.y = spval(obj.ps_y,t);
            obj.dy = spval(obj.ps_dy,t);
            obj.ddy = spval(obj.ps_ddy,t);

        end
        
        function f = plot(obj,t)
           eval(obj,t);
           
           f=figure;
           plot(t,obj.y,t,obj.dy,t,obj.ddy);
           title('polynomial trajectory')
           grid on;
           legend('$y$','$\dot{y}$','$\ddot{y}$','Interpreter','latex');
        end
        
        function f = plot_spline(obj,t)
           eval_spline(obj,t);
           
           f=figure;
           plot(t,obj.y,t,obj.dy,t,obj.ddy);
           title('spline trajectory')
           grid on;
           legend('$y$','$\dot{y}$','$\ddot{y}$','Interpreter','latex');
        end
    end
end

