classdef KinematicCarReference < KinematicCarControl
    %KINEMATICCARREFERENCE is an open-loop controller for the kinematic car
    %based on the flat output trajectory t\mapsto (y1(t),y2(t)) with
    %components expressed as polynomials in time
    %See Flatness-based Control: An Introduction, Shaker Verlag, Feb. 2021
    %section 1.7, p.10-14
    
    properties (SetAccess=public, GetAccess=public)
        y1;
        y2;
        dy1;
        dy2;
        ddy1;
        ddy2;
        th;
    end
    
    properties (SetAccess=private, GetAccess=public)
        l (1,1) double {mustBePositive} = 1; % wheelbase length
    end
    properties (SetAccess=private, GetAccess=private)
        r1 PolyRef = PolyRef(); % reference to trajectory for t\mapsto y1(t)
        r2 PolyRef = PolyRef(); % reference to trajectory for t\mapsto y2(t)
    end
    
    methods
        function obj = KinematicCarReference(wheelbase,y1r,y2r)
            %KINEMATICCARCONTROL Construct an instance of this class
            if  nargin > 0
                obj.l = wheelbase;
                obj.r1 = y1r;
                obj.r2 = y2r;
            end
        end
        
        function obj = eval(obj,t)
            %EVAL evaluates the control variables y1,y2,th,v,phi given the
            %time vector t and stores the result in the object properties
            
            % evaluate the reference trajectory
            eval(obj.r1,t);
            eval(obj.r2,t);
            
            % get the reference trajectory variables for easier readability
            obj.y1 = obj.r1.y;
            obj.y2 = obj.r2.y;
            obj.dy1 = obj.r1.dy;
            obj.dy2 = obj.r2.dy;
            obj.ddy1 = obj.r1.ddy;
            obj.ddy2 = obj.r2.ddy;
            
            % open loop controller (1.9)     
            obj.v = sqrt((obj.dy1).^2 + (obj.dy2).^2);
            obj.phi = atan(obj.l*(obj.ddy2.*obj.dy1-obj.ddy1.*obj.dy2)./(obj.v).^3);
            obj.th = atan(obj.dy2./obj.dy1);
        end
        
    end
end

