classdef KinematicCarReference < KinematicCarControl
    %KINEMATICCARREFERENCE is an open-loop controller for the kinematic car
    %based on the explicit path parametrization y2 = f(y1) for points 
    %(y1,y2) in the plane.
    %See Flatness-based Control: An Introduction, Shaker Verlag, Feb. 2021
    %section 3.3.6, p.79-83
    
    properties (SetAccess=public, GetAccess=public)
        y1;
        y2;
        th;
        kappa;
        kappap;        
    end
    
    properties (SetAccess=private, GetAccess=public)
        l (1,1) double {mustBePositive} = 1; % wheelbase length
        t_star = 5;          % trajectory end time
        yA (2,1)double;      % planned initial position
        thetaA (1,1) double; % planned initial orientation
        yB (2,1)double;      % planned final position
        thetaB (1,1) double; % planned final orientation
        a = zeros(4,1);     
        b = zeros(2,1);
        RA = zeros(2);
        tyB = zeros(2,1);
        tthetaB = 0;
    end
    
    methods
        function obj = KinematicCarReference(wheelbase,t_star,yA,thetaA,phiA,yB,thetaB,phiB)
            %KINEMATICCARCONTROL Construct an instance of this class
            if  nargin > 0
                obj.l = wheelbase;
                obj.yA = yA;
                obj.yB = yB;
                obj.thetaA = thetaA;
                obj.thetaB = thetaB;
                obj.t_star = t_star;
            end
            
            % save some parameters for later transformations
            %\tilde{\theta}_B = \theta_B - \theta_A
            obj.tthetaB = thetaB - thetaA;
            % rotation matrix R(\theta_A)
            obj.RA = [cos(obj.thetaA) -sin(obj.thetaA); sin(obj.thetaA) cos(obj.thetaA)];
            %\tilde{y}_B = R(\theta_A)*(y_B-y_A)
            obj.tyB=obj.RA'*(obj.yB-obj.yA);
            
            % calculate polynomial coefficients
            obj.calcCoeff(phiA,phiB);
        end
                  
        function obj = calcCoeff(obj,phiA,phiB)
            % calculate coefficients of transformed polynomial \tilde{y}
            % where y = yA + RA*\tilde{y}
            % or \tilde{y} = RA'*(y-yA)
            % with \tilde(theta) = theta - thetaA
            % see 3.3.6., p. 81
            obj.a(1) = obj.tyB(1)^2*tan(phiA)/2/obj.l;
            obj.a(2:4) = 1/2/obj.l*inv([1 1 1;3 4 5;3 6 10])*...
                [2*obj.l*obj.tyB(2) - obj.tyB(1)^2*tan(phiA);...
                2*obj.l*obj.tyB(1)*tan(obj.tthetaB) - 2*obj.tyB(1)^2*tan(phiA);...
                obj.tyB(1)^2*tan(phiB)*(1+(tan(obj.tthetaB))^2)^(3/2) - obj.tyB(1)^2*tan(phiA)];
            obj.b = [3;-2];
        end
        
        function out = eval(obj,tt)
            %EVAL evaluates the nominal control input components (v,phi) 
            %given the time vector t.
            %The result is stored in the object properties and in the
            %output struct out.
            
            % saturate time signal to interval [t,t_star] to hold
            % trajectory at its endpoints
            
            if tt<0
                t = 0; 
            elseif tt>obj.t_star
                t = obj.t_star;
            else
                t = tt;
            end
            
            % evaluate polynomial \tilde{y}
            ty1 = obj.tyB(1)*obj.b'*[(t/obj.t_star)^2;(t/obj.t_star)^3];
            ty2 = obj.a'*[(ty1/obj.tyB(1))^2;(ty1/obj.tyB(1))^3;(ty1/obj.tyB(1))^4;(ty1/obj.tyB(1))^5];
            
            % calculate derivatives
            %dty1/dt
            dty1_dt = obj.tyB(1)*obj.b'*[2*t/(obj.t_star)^2;3*t^2/(obj.t_star)^3];
            
            %d\tilde{y2}/d\tilde{y1}
            dty2_dty1 = obj.a'*[2*ty1/(obj.tyB(1))^2;3*(ty1)^2/(obj.tyB(1))^3;4*ty1^3/(obj.tyB(1))^4;5*ty1^4/(obj.tyB(1))^5];
            %d^2\tilde{y2}/d\tilde{y1}^2
            d2ty2_dty1 = obj.a'*[2/(obj.tyB(1))^2;6*(ty1)/(obj.tyB(1))^3;12*ty1^2/(obj.tyB(1))^4;20*ty1^3/(obj.tyB(1))^5];
            %d^3\tilde{y2}/d\tilde{y1}^3
            d3ty2_dty1 = obj.a'*[0;6/(obj.tyB(1))^3;24*ty1/(obj.tyB(1))^4;60*ty1^2/(obj.tyB(1))^5];
            
            %d\tilde{y1}/d\tilde{s} = \tilde{y1}'
            ty1p = 1/sqrt(1+dty2_dty1^2);
            %d\tilde{y2}/d\tilde{s} = \tilde{y2}'
            ty2p = dty2_dty1*ty1p;
            
            % cos(\tilde{theta})
            c_ttheta = ty1p/sqrt(ty1p^2 + ty2p^2);
            % sin(\tilde{theta})
            s_ttheta = ty2p/sqrt(ty1p^2 + ty2p^2);
            ttheta = atan(s_ttheta/c_ttheta);
            obj.th = ttheta + obj.thetaA;
            
            %transform to y, where y = yA + RA*\tilde{y}
            y = obj.yA + obj.RA*[ty1;ty2];
            obj.y1 = y(1);obj.y2 = y(2);
            yp = obj.RA*[ty1p;ty2p];
            
            % cos(theta)
            c_theta = yp(1)/sqrt(yp(1)^2 + yp(2)^2);
            % sin(theta)
            s_theta = yp(2)/sqrt(yp(1)^2 + yp(2)^2);
            theta = atan(s_theta/c_theta);
            
            % transform derivatives
            c_thetaA = cos(obj.thetaA);
            s_thetaA = sin(obj.thetaA);
            dy1_dty1 = c_thetaA - s_thetaA*dty2_dty1;
            dy2_dty1 = s_thetaA + c_thetaA*dty2_dty1;

            d2y1_dty1 = -s_thetaA*d2ty2_dty1;
            d2y2_dty1 = c_thetaA*d2ty2_dty1;
            d3y1_dty1 = -s_thetaA*d3ty2_dty1;
            d3y2_dty1 = c_thetaA*d3ty2_dty1;
            
            dy1_dt = dy1_dty1*dty1_dt;
            dy2_dy1 = dy2_dty1/dy1_dty1;
            d2y2_dy1 = (d2y2_dty1-dy2_dy1*d2y1_dty1)/dy1_dty1^2;
            d3y2_dy1 = (d3y2_dty1 - dy2_dy1*d3y1_dty1 - dy1_dty1*d2y2_dy1*d2y1_dty1)/dy1_dty1^3 - 2*(d2y2_dty1 -dy2_dy1*d2y1_dty1)*d2y1_dty1/dy1_dty1^4;
            
            % curvature \kappa, eq. (3.39)
            obj.kappa = d2y2_dy1/(1+dy2_dy1^2)^(3/2);
            % d\kappa/ds
            obj.kappap = yp(1)^4*d3y2_dy1 - 3*obj.kappa^2*dy2_dy1;
            
            % vr = \dot{sr} = dy1/dt*dsr/dy1 = dy1/dt*(y1')^(-1)
            % since dsr/dy1 = sqrt(1+(dy2/dy1)^2) = 1/y1'
            obj.v = dy1_dt/yp(1);
 
            % steering angle
            obj.phi = sign(obj.v)*atan(obj.l*obj.kappa);
            
            % output variables for plotting
            out.y1 = y(1);
            out.y2 = y(2);
            out.th = obj.th;
            out.v = obj.v;
            out.phi = obj.phi;
            out.kappa = obj.kappa;
        end    
    end
end

