classdef KinematicCarQuasistaticControl < KinematicCarClosedLoopControl
    %KINEMATICCARQUASISTATICCONTROL is an quasi-static controller for the kinematic car
    %using a moving frame 
    %See Flatness-based Control: An Introduction, Shaker Verlag, Feb. 2021
    %section 4.5.4, p.116-123
    
    properties (SetAccess=private, GetAccess=public)
        ktau0 = 1;
        knu0 = 1;
        knu1 = 1;
    end
       
    methods
        function obj = KinematicCarQuasistaticControl(ktau0, knu0, knu1 , r)
            %KINEMATICCARQUASISTATICCONTROL Construct an instance of this class
            if  nargin > 0
                % save controller gains
                obj.ktau0 = ktau0;
                obj.knu0 = knu0;
                obj.knu1 = knu1;
                obj.r = r;
            end
        end
        
        % component tangential to path
        function tau = calc_tau(obj,theta)
            tau = [cos(theta); sin(theta)];
        end
        
        % component normal to path
        function nu = calc_nu(obj,theta)
            nu = [-sin(theta); cos(theta)];
        end
                          
        function out = eval(obj,t,y1,y2,theta)
            %EVAL evaluates the control input components (v,phi) given the
            %time vector t, state (y1, y2, theta), and reference r
            %The result is stored in the object properties and in the
            %output struct out.
        
            % compute system inputs and internal variables 
            % for quasi-static controller (see sec. 4.5.4. p. 116-123)

                        
            %% evaluate the reference
            eval(obj.r,t);
            r=obj.r;
            
            %% compute quasi-static controller
            tau_r = obj.calc_tau(r.th);
            nu_r = obj.calc_nu(r.th);
            
            % project current position on reference axes
            ytau = [y1 y2]*tau_r;
            ynu = [y1 y2]*nu_r;
            % get curent heading error wrt to reference heading
            tau = obj.calc_tau(theta);
            c_delta = tau'*tau_r;
            s_delta = tau'*nu_r;
            
            % compute reference variables
            % eqn. (4.80d)-(4.80i)
            ytau_r=[r.y1 r.y2]*tau_r;
            ynu_r=[r.y1 r.y2]*nu_r;
            ytaup_r = 1+r.kappa*ynu_r;
            ynup_r = -r.kappa*ytau_r;
            ytaupp_r = r.kappap*ynu_r -r.kappa^2*ytau_r;
            ynupp_r = -r.kappap*ytau_r -r.kappa*(1+r.kappa*ynu_r);
            
            % feedback in tau direction
            % eqn. (4.80a)
            wtau = ytaup_r - sign(r.v)*obj.ktau0*(ytau-ytau_r);
            % transform to physical input v
            % eqn. (4.78)
            u = (wtau - r.kappa*ynu)/c_delta;
            obj.v = u*r.v;

            % feedback in nu direction
            ynup = u*s_delta - r.kappa*ytau;
            % eqn. (4.80b)
            wtaup = ytaupp_r - sign(r.v)*obj.ktau0*(wtau-ytaup_r);
            % eqn. (4.80c)
            wnu = ynupp_r - sign(r.v)*obj.knu1*(ynup-ynup_r) - obj.knu0*(ynu-ynu_r);
            % transform to physical input kappa
            % second row of (4.79)
            kappa=1/u*r.kappa+1/u^2*[-s_delta c_delta]*([-ynu -ynup;ytau wtau]*[r.kappap;r.kappa]+[wtaup;wnu]);
            
            % steering angle
            obj.phi = atan(r.l*kappa);
            
            %% output physical system inputs
            out.v = obj.v;
            out.phi = obj.phi;
            out.kappa = kappa;
            
        end
    end
end

